@@ -17,8 +17,7 @@
 """Nessie Data objects."""
 import re
 from datetime import datetime
-from typing import List
-from typing import Optional
+from typing import List, Optional, Tuple
 
 import attr
 import desert
@@ -26,6 +25,39 @@ from marshmallow import fields
 from marshmallow_oneofschema import OneOfSchema
 
 
+# regex taken from org.projectnessie.model.Validation
+__RE_REFERENCE_NAME_RAW = "[A-Za-z](((?![.][.])[A-Za-z0-9./_-])*[A-Za-z0-9._-])"
+__RE_REFERENCE_NAME: re.Pattern = re.compile(f"^{__RE_REFERENCE_NAME_RAW}$")
+__RE_HASH_RAW = "[0-9a-fA-F]{8,64}"
+__RE_HASH: re.Pattern = re.compile(f"^{__RE_HASH_RAW}$")
+__RE_REFERENCE_WITH_HASH: re.Pattern = re.compile(f"^({__RE_REFERENCE_NAME_RAW})([@]|[*])({__RE_HASH_RAW})$")
+
+
+DETACHED_REFERENCE_NAME = "DETACHED"
+
+
+def is_valid_reference_name(ref: str) -> bool:
+    """Checks whether 'ref' is a valid reference name."""
+    return __RE_REFERENCE_NAME.match(ref) is not None
+
+
+def is_valid_hash(ref: str) -> bool:
+    """Checks whether 'ref' is a valid commit id/hash."""
+    return __RE_HASH.match(ref) is not None
+
+
+def split_into_reference_and_hash(ref_with_hash: Optional[str]) -> Tuple[str, Optional[str]]:
+    """Returns a tuple of reference-name + hash, if the given string represents a ref-name + hash tuple 'ref_name@commit_id'."""
+    if not ref_with_hash:
+        return "<UNKNOWN>", None
+    match = __RE_REFERENCE_WITH_HASH.match(ref_with_hash)
+    if not match:
+        if is_valid_hash(ref_with_hash):
+            return DETACHED_REFERENCE_NAME, ref_with_hash
+        return ref_with_hash, None
+    return match.group(1), match.group(5)
+
+
 @attr.dataclass
 class Content:
     """Dataclass for Nessie Content."""
@@ -90,18 +122,23 @@ DeltaLakeTableSchema = desert.schema_class(DeltaLakeTable)
 
 
 @attr.dataclass
-class SqlView(Content):
-    """Dataclass for Nessie SQL View."""
+class IcebergView(Content):
+    """Dataclass for Nessie Iceberg View."""
 
+    metadata_location: str = desert.ib(fields.Str(data_key="metadataLocation"))
+    version_id: int = desert.ib(fields.Int(data_key="versionId"))
+    schema_id: int = desert.ib(fields.Int(data_key="schemaId"))
     dialect: str = desert.ib(fields.Str())
     sql_text: str = desert.ib(fields.Str(data_key="sqlText"))
 
-    def pretty_print(self: "SqlView") -> str:
+    def pretty_print(self: "IcebergView") -> str:
         """Print out for cli."""
-        return "SqlView:\n\tDialect: {}\n\tSql: {}".format(self.dialect, self.sql_text)  # todo use a sql parser to pretty print this
+        return "IcebergView:\n\tmetadata-location: {}\n\tversion-id: {}\n\tschema-id: {}\n\tDialect: {}\n\tSql: {}".format(
+            self.metadata_location, self.version_id, self.schema_id, self.dialect, self.sql_text
+        )  # todo use a sql parser to pretty print this
 
 
-SqlViewSchema = desert.schema_class(SqlView)
+IcebergViewSchema = desert.schema_class(IcebergView)
 
 
 class ContentSchema(OneOfSchema):
@@ -110,7 +147,7 @@ class ContentSchema(OneOfSchema):
     type_schemas = {
         "ICEBERG_TABLE": IcebergTableSchema,
         "DELTA_LAKE_TABLE": DeltaLakeTableSchema,
-        "VIEW": SqlViewSchema,
+        "ICEBERG_VIEW": IcebergViewSchema,
     }
 
     def get_obj_type(self: "ContentSchema", obj: Content) -> str:
@@ -119,8 +156,8 @@ class ContentSchema(OneOfSchema):
             return "ICEBERG_TABLE"
         if isinstance(obj, DeltaLakeTable):
             return "DELTA_LAKE_TABLE"
-        if isinstance(obj, SqlView):
-            return "VIEW"
+        if isinstance(obj, IcebergView):
+            return "ICEBERG_VIEW"
 
         raise Exception("Unknown object type: {}".format(obj.__class__.__name__))
 
@@ -280,6 +317,16 @@ class Reference:
     )
 
 
+@attr.dataclass
+class Detached(Reference):
+    """Dataclass for Nessie detached commit id."""
+
+    pass
+
+
+DetachedSchema = desert.schema_class(Detached)
+
+
 @attr.dataclass
 class Branch(Reference):
     """Dataclass for Nessie Branch."""
@@ -306,6 +353,7 @@ class ReferenceSchema(OneOfSchema):
     type_schemas = {
         "BRANCH": BranchSchema,
         "TAG": TagSchema,
+        "DETACHED": DetachedSchema,
     }
 
     def get_obj_type(self: "ReferenceSchema", obj: Reference) -> str:
@@ -314,6 +362,8 @@ class ReferenceSchema(OneOfSchema):
             return "BRANCH"
         if isinstance(obj, Tag):
             return "TAG"
+        if isinstance(obj, Detached):
+            return "DETACHED"
 
         raise Exception("Unknown object type: {}".format(obj.__class__.__name__))
 
